
/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "cart_state_estimator/cart_state_estimator.h"
#include "cart_state_estimator/util.h"

namespace cart_state_estimator
{
const string CART_FRAME("cart");
const string BASE_FRAME("base_footprint");
const string CHECKERBOARD_FRAME("CRT");

/************************************************************
 * Initialization
 ************************************************************/

gm::PolygonStamped computeFootprint (const double w, const double l)
{
  gm::PolygonStamped footprint;
  footprint.header.frame_id = CART_FRAME;
  footprint.polygon.points.resize(4);
  footprint.polygon.points[1].x = l;
  footprint.polygon.points[2].x = l;
  footprint.polygon.points[2].y = w;
  footprint.polygon.points[3].y = w;
  return footprint;
}

string toString (const btTransform& tr)
{
  btVector3 t = tr.getOrigin();
  btQuaternion r = tr.getRotation();
  return (format("((%.3f, %.3f, %.3f), (%.3f, %.3f, %.3f, %.3f))") %
          t.x() % t.y() % t.z() % r.x() % r.y() % r.z() % r.w()).str();
}

CartStateEstimator::CartStateEstimator () :
  cart_width_(getParam<double>("width", "cart_pushing")),
  cart_length_(getParam<double>("length", "cart_pushing")),
  l_gripper_frame_(getParam<string>("l_gripper_frame")),
  r_gripper_frame_(getParam<string>("r_gripper_frame")),
  publish_fake_transforms_(getPrivateParam<bool>("publish_fake_gripper_transforms", false)),
  footprint_(computeFootprint(cart_width_, cart_length_)),
  footprint_x_offset_(getParam<double>("footprint_x_offset", "cart_pushing")),
  footprint_y_offset_(getParam<double>("footprint_y_offset", "cart_pushing")),
  publication_interval_(getPrivateParam<double>("publication_interval", 0.05)),
  transform_recency_threshold_(getPrivateParam<double>("transform_recency_threshold", 0.4)),
  cart_to_board_(readTransformParameter(string("cart_pushing/checkerboard_pose")).inverse()),
  pub_timer_(nh_.createWallTimer(WallDuration(publication_interval_), &CartStateEstimator::publishState, this)),
  footprint_pub_(nh_.advertise<gm::PolygonStamped> ("cart_footprint", 10)),
  cart_init_pose_(readTransformParameter(string("cart_init_pose")))
{
    reset_est_server_ = nh_.advertiseService("cart_pushing/reset_estimate", &CartStateEstimator::resetEstimate, this);

    /* Uncomment to use local grasp_solver_ call rather than service call */
	gripper_init_poses_.push_back(readTransformParameter(string("r_gripper_grasp")));
	gripper_init_poses_.push_back(readTransformParameter(string("l_gripper_grasp")));
	grasp_solver_ = ManipulationTransforms(cart_init_pose_, gripper_init_poses_);
	ROS_ASSERT (cart_width_ > 0.0 && cart_length_ > 0.0);
}

bool CartStateEstimator::resetEstimate(std_srvs::Empty::Request &req, std_srvs::Empty::Response &resp) {
    ROS_INFO("Cart position estimate reset by service call");
    Lock l(mutex_);
    tf::StampedTransform xfm1, xfm2;
    try {
        tf_listener_.lookupTransform (BASE_FRAME, l_gripper_frame_, ros::Time(), xfm1);
        tf_listener_.lookupTransform (BASE_FRAME, r_gripper_frame_, ros::Time(), xfm2);
    }
    catch (tf::TransformException ex) {
        ROS_ERROR("Transform exception attempting to set initial cart pose: %s", ex.what());
        return false;
    }

    grasp_solver_.setInitialTransforms(cart_init_pose_, xfm1, xfm2);
    return true;
}


// Returns cart polygon in base footprint frame
gm::PolygonStamped CartStateEstimator::getProjectedFootprint () const 
{
  gm::PolygonStamped footprint = footprint_;
  tf::StampedTransform cart_to_base;
  tf_listener_.lookupTransform (BASE_FRAME, CART_FRAME, Time(), cart_to_base);
  for (unsigned i = 0; i < 4; i++) {
		btVector3 p(footprint.polygon.points[i].x - footprint_x_offset_,
				footprint.polygon.points[i].y - footprint_y_offset_,
				footprint.polygon.points[i].z);
		btVector3 trans = cart_to_base * p;
		footprint.polygon.points[i].x = trans.x();
		footprint.polygon.points[i].y = trans.y();
		footprint.polygon.points[i].z = 0.0;
	}
  footprint.header.stamp = Time::now();
  footprint.header.frame_id = BASE_FRAME;
  return footprint;
}


void CartStateEstimator::broadcastCartPose(const string& frame, const btTransform& trans)
{
  tf::StampedTransform stamped(trans, ros::Time::now(), BASE_FRAME, frame);
  ROS_DEBUG_STREAM_NAMED("cart_state_estimator", "broadcasting cart pose: " << toString(trans));
  tf_broadcaster_.sendTransform(stamped);
}

void CartStateEstimator::publishState (const WallTimerEvent& e) 
{
  try {
    footprint_pub_.publish(getProjectedFootprint());
  }
  catch (tf::TransformException &ex) {
    ROS_DEBUG_STREAM_NAMED ("transforms", "Not publishing footprint due to transform exception " << ex.what());
  }

  //ROS_DEBUG_NAMED ("node", "Acquiring lock");
  Lock l(mutex_);

  /****************************************
   * Use gripper positions to get
   * proprioceptive estimate
   ****************************************/

  optional<btTransform> proprioceptive_estimate;
  tf::StampedTransform l_gripper_to_base, r_gripper_to_base;
  /* Uncomment to use local grasp_solver_ call rather than service call */
  std::vector<tf::Transform> gripper_tfT_base; gripper_tfT_base.resize(2); // Can't be StampedTransform b/c vector clobbers type, and can't be passed
  std::vector<gm::PoseStamped> gripper_psT_base; gripper_psT_base.resize(2);
  const Time most_recent;

  Time cutoff;
  try {
    cutoff = Time::now() - Duration(transform_recency_threshold_);
  } catch (std::runtime_error &e){
    cutoff = Time::now();
  }

  try {
      // Find gripper poses in base link
    tf_listener_.lookupTransform (BASE_FRAME, l_gripper_frame_, most_recent, l_gripper_to_base);
    tf_listener_.lookupTransform (BASE_FRAME, r_gripper_frame_, most_recent, r_gripper_to_base);
    if (l_gripper_to_base.stamp_ > cutoff && r_gripper_to_base.stamp_ > cutoff) {
      btTransform estimate;
      ROS_DEBUG_STREAM ("R gripper pose is " << toString(r_gripper_to_base) <<
                       " l gripper is " << toString(l_gripper_to_base));

      // Compute object pose using solver directly
      /* Uncomment to use local grasp_solver_ call rather than service call */
      gripper_tfT_base[0] = btTransform(r_gripper_to_base);
      gripper_tfT_base[1] = btTransform(l_gripper_to_base);
      grasp_solver_.mapEffectorPosesToObject(gripper_tfT_base, estimate);

      proprioceptive_estimate = estimate;
      broadcastCartPose("cart_proprioceptive", *proprioceptive_estimate);
    }
    else
      ROS_DEBUG_STREAM_NAMED ("transforms", "Transforms at " << l_gripper_to_base.stamp_ <<
                              " and " << r_gripper_to_base.stamp_ << " are not within cutoff " << cutoff);
  }
  catch (tf::TransformException &ex) {
    ROS_DEBUG_STREAM_NAMED ("transforms", "Skipping proprioceptive estimate due to transform error " << ex.what());
  }

  if (proprioceptive_estimate) {
    ROS_DEBUG_NAMED ("cart_pose_estimate", "Using proprioceptive estimate");
    broadcastCartPose(CART_FRAME, *proprioceptive_estimate);
  }
  else
    ROS_DEBUG_NAMED ("transforms", "Not publishing cart pose as proprioceptive estimate was available");
}

} // namespace cart_state_estimator

int main (int argc, char** argv)
{
  ros::init(argc, argv, "cart_state_estimator");
  cart_state_estimator::CartStateEstimator node;
  ros::MultiThreadedSpinner spinner(2); // Because the pub thread sometimes waits for the fake pub thread
  spinner.spin();
  return 0;
}
